#include <gtk/gtk.h>
#include <sys/time.h>
#include <time.h>
#include <stdlib.h>

#include <dbus/dbus-glib.h>

#ifndef GTKDEBUG
#include <mce/dbus-names.h>
#include <mce/mode-names.h>
#include <libhildondesktop/libhildondesktop.h>
#endif

#include "datasource.h"
#include "mh-home-sysmon.h"

#include "mh-cpu-source.h"
#include "mh-mem-source.h"

#include <stdio.h>
#include <string.h>

#define MH_HOME_SYSMON_GET_PRIVATE(x) (G_TYPE_INSTANCE_GET_PRIVATE((x), MH_TYPE_HOME_SYSMON, MhHomeSysmonPrivate))

typedef struct _MhHomeSysmonPrivate MhHomeSysmonPrivate;
struct _MhHomeSysmonPrivate {
	GQueue *line_data;
	GQueue *solid_data;

#ifndef GTKDEBUG
	DBusGConnection *gconn;
	DBusGProxy *proxy;
#endif

	DataSource *line_source;
	DataSource *solid_source;

	PangoFontDescription *desc;

	guint datapoints;
	guint timer_source;

	gboolean foreground;
};

#define TEXT_ALIGN_VMASK 0x3
#define TEXT_ALIGN_HMASK 0xc
typedef enum _TextAlign TextAlign;
enum _TextAlign {
	TEXT_ALIGN_DEFAULT = 0,
	TEXT_ALIGN_TOP = 1,
	TEXT_ALIGN_BOTTOM = 2,
	TEXT_ALIGN_MIDDLE = 3,
	TEXT_ALIGN_LEFT = 1 << 2,
	TEXT_ALIGN_RIGHT = 2 << 2,
	TEXT_ALIGN_CENTER = 3 << 2
};

static gboolean mh_home_sysmon_expose(GtkWidget *widget,
				      GdkEventExpose *event);
static void mh_home_sysmon_screen_changed(GtkWidget *widget,
					  GdkScreen *old);
static gboolean mh_home_sysmon_update(gpointer user_data);
static void cairo_draw_text(cairo_t * cr,
			    PangoFontDescription * desc,
			    GtkRequisition *req,
			    gint padding,
			    const gchar *markup, TextAlign align);
static void mh_home_sysmon_foreground(HildonDesktopHomeItem *item);
static void mh_home_sysmon_background(HildonDesktopHomeItem *item);
static void mh_home_sysmon_display_ind(DBusGProxy *proxy,
				       gchar *display_state,
				       MhHomeSysmon *sysmon);

#ifndef GTKDEBUG
HD_DEFINE_PLUGIN(MhHomeSysmon, mh_home_sysmon,
		 HILDON_DESKTOP_TYPE_HOME_ITEM);
#else
G_DEFINE_TYPE(MhHomeSysmon, mh_home_sysmon, GTK_TYPE_WINDOW);
#endif

static void mh_home_sysmon_finalize(GObject *obj)
{
	MhHomeSysmonPrivate *priv = MH_HOME_SYSMON_GET_PRIVATE(obj);
	g_source_remove(priv->timer_source);
	g_queue_free(priv->line_data);
	g_queue_free(priv->solid_data);

	g_object_unref(priv->proxy);
	g_object_unref(priv->gconn);
}

static void mh_home_sysmon_init(MhHomeSysmon * plugin)
{
	MhHomeSysmonPrivate *priv = MH_HOME_SYSMON_GET_PRIVATE(plugin);
	priv->line_data = g_queue_new();
	priv->solid_data = g_queue_new();

	priv->line_source = g_new0(DataSource, 1);
	priv->solid_source = g_new0(DataSource, 1);

	mh_cpu_source_load(priv->line_source);
	mh_mem_source_load(priv->solid_source);

	priv->line_source->ds_data =
		priv->line_source->init(priv->line_source);
	priv->solid_source->ds_data =
		priv->solid_source->init(priv->solid_source);

	priv->datapoints = 20;

	priv->desc = pango_font_description_from_string("Sans 7");

	priv->timer_source =
		g_timeout_add(5000, mh_home_sysmon_update, plugin);

	priv->foreground = TRUE;

	gtk_widget_set_size_request(GTK_WIDGET(plugin), 140, 140);
#ifndef GTKDEBUG
	hildon_desktop_home_item_set_resize_type(HILDON_DESKTOP_HOME_ITEM
						 (plugin),
						 HILDON_DESKTOP_HOME_ITEM_RESIZE_BOTH);

	priv->gconn = dbus_g_bus_get(DBUS_BUS_SYSTEM, NULL);
	priv->proxy = dbus_g_proxy_new_for_name(priv->gconn,
						MCE_SERVICE,
						MCE_SIGNAL_PATH,
						MCE_SIGNAL_IF);
	dbus_g_proxy_add_signal(priv->proxy,
				MCE_DISPLAY_SIG,
				G_TYPE_STRING,
				G_TYPE_INVALID);
	dbus_g_proxy_connect_signal(priv->proxy,
				    MCE_DISPLAY_SIG,
				    G_CALLBACK(mh_home_sysmon_display_ind),
				    plugin,
				    NULL);
#endif

	gtk_widget_show_all(GTK_WIDGET(plugin));
}

static void mh_home_sysmon_class_init(MhHomeSysmonClass * klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS(klass);
	GtkWidgetClass *widget_class = GTK_WIDGET_CLASS(klass);
	HildonDesktopHomeItemClass *item_class =
		HILDON_DESKTOP_HOME_ITEM_CLASS(klass);
	g_type_class_add_private(klass, sizeof(MhHomeSysmonPrivate));

	object_class->finalize = mh_home_sysmon_finalize;

	widget_class->expose_event = mh_home_sysmon_expose;
	widget_class->screen_changed = mh_home_sysmon_screen_changed;

	item_class->background = mh_home_sysmon_background;
	item_class->foreground = mh_home_sysmon_foreground;
}

static gboolean mh_home_sysmon_update(gpointer user_data)
{
	MhHomeSysmon *plugin = user_data;
	MhHomeSysmonPrivate *priv = MH_HOME_SYSMON_GET_PRIVATE(plugin);

	if (priv->line_source) {
		guint val =
			priv->line_source->get_value(priv->line_source->
						     ds_data) * UINT_MAX;
		g_queue_push_head(priv->line_data, GUINT_TO_POINTER(val));

	}
	if (priv->solid_source) {
		guint val =
			priv->solid_source->get_value(priv->solid_source->
						      ds_data) * UINT_MAX;
		g_queue_push_head(priv->solid_data, GUINT_TO_POINTER(val));

	}

	while (g_queue_get_length(priv->line_data) > priv->datapoints) {
		g_queue_pop_tail(priv->line_data);
	}
	if (g_queue_get_length(priv->solid_data) > priv->datapoints) {
		g_queue_pop_tail(priv->solid_data);
	}

	if (!priv->foreground) {
		gtk_widget_queue_draw(GTK_WIDGET(plugin));
	}

	return TRUE;
}

#ifndef PI
#define PI 3.1415926535
#endif
static gboolean mh_home_sysmon_expose(GtkWidget *widget,
				      GdkEventExpose *event)
{
	cairo_t *cr;
	GtkRequisition req;
	cairo_path_t *path;

	guint i;
	GList *list;

	MhHomeSysmonPrivate *priv = MH_HOME_SYSMON_GET_PRIVATE(widget);

	cairo_pattern_t *pattern;

	(void)event;

	cr = gdk_cairo_create(GDK_DRAWABLE(widget->window));
	req.height = widget->allocation.height;
	req.width = widget->allocation.width;

	pattern = cairo_pattern_create_linear(0, 0, 0, req.height);
	cairo_pattern_add_color_stop_rgba(pattern, 0, 1.0, 1.0, 1.0, 0.5);
	cairo_pattern_add_color_stop_rgba(pattern, 0.58579, 1.0, 1.0, 1.0,
					  0.2);
	cairo_pattern_add_color_stop_rgba(pattern, 0.58579, 0, 0, 0, 0.4);
	cairo_pattern_add_color_stop_rgba(pattern, 1, 1.0, 1.0, 1.0, 0.5);

	cairo_set_source_rgba(cr, 0.0, 0.0, 0.0, 0.0);
	cairo_set_operator(cr, CAIRO_OPERATOR_SOURCE);
	cairo_paint(cr);

	cairo_set_operator(cr, CAIRO_OPERATOR_OVER);
	cairo_move_to(cr, 15.5, 5.5);
	cairo_line_to(cr, req.width - 15.5, 5.5);
	cairo_arc(cr, req.width - 15.5, 15.5, 10, -PI / 2, 0);
	cairo_line_to(cr, req.width - 5.5, req.height - 15.5);
	cairo_arc(cr, req.width - 15.5, req.height - 15.5, 10, 0, PI / 2);
	cairo_line_to(cr, 15.5, req.height - 5.5);
	cairo_arc(cr, 15.5, req.height - 15.5, 10, PI / 2, PI);
	cairo_line_to(cr, 5.5, 15.5);
	cairo_arc(cr, 15.5, 15.5, 10, -PI, -PI / 2);

	cairo_set_line_width(cr, 2.0);
	cairo_set_source(cr, pattern);
	cairo_fill_preserve(cr);
	path = cairo_copy_path(cr);
	cairo_clip(cr);

	cairo_pattern_destroy(pattern);

	cairo_move_to(cr, (req.width - 7.5), (req.height - 7.5));
	for (i = 0, list = g_queue_peek_head_link(priv->solid_data);
	     i < priv->datapoints && list; i++, list = list->next) {
		cairo_line_to(cr,
			      ((req.width - 14) * (priv->datapoints - i -
						   1) / (priv->datapoints -
							 1)) + 7.5,
			      ((UINT_MAX - GPOINTER_TO_UINT(list->data)) *
			       (req.height - 14.0) / UINT_MAX) + 7.5);
	}
	cairo_line_to(cr,
		      ((req.width - 14) * (priv->datapoints -
					   i) / (priv->datapoints - 1)) +
		      7.5, (req.height - 7.5));
	cairo_set_source_rgba(cr, 0.0, 0.0, 1.0, 0.7);
	cairo_fill(cr);

	for (i = 0, list = g_queue_peek_head_link(priv->line_data);
	     i < priv->datapoints && list; i++, list = list->next) {
		cairo_line_to(cr,
			      ((req.width - 14) * (priv->datapoints - i -
						   1) / (priv->datapoints -
							 1)) + 7.5,
			      ((UINT_MAX - GPOINTER_TO_UINT(list->data)) *
			       (req.height - 14.0) / UINT_MAX) + 7.5);
	}
	cairo_set_source_rgba(cr, 1.0, 0, 0, 0.9);
	cairo_stroke(cr);

	cairo_append_path(cr, path);
	cairo_set_source_rgba(cr, 1.0, 1.0, 1.0, 0.9);
	cairo_stroke(cr);
	cairo_path_destroy(path);

	cairo_set_source_rgba(cr, 0.0, 0.0, 0.0, 1.0);
#define TEXT(src, text, valign, halign) \
	if (priv->src ## _source && priv->src ## _source->text ## _text) \
		cairo_draw_text(cr, priv->desc, &req, 9, priv->src ## _source->text ## _text, TEXT_ALIGN_ ## valign | TEXT_ALIGN_ ## halign);
	TEXT(line, max, TOP, LEFT);
	TEXT(line, half, MIDDLE, LEFT);
	TEXT(line, min, BOTTOM, LEFT);
	TEXT(solid, max, TOP, RIGHT);
	TEXT(solid, half, MIDDLE, RIGHT);
	TEXT(solid, min, BOTTOM, RIGHT);
#undef TEXT

	cairo_destroy(cr);

	return TRUE;
}

static void mh_home_sysmon_screen_changed(GtkWidget *widget,
					  GdkScreen *old)
{
	GdkScreen *screen = gtk_widget_get_screen(widget);
	GdkColormap *cmap = gdk_screen_get_rgba_colormap(screen);

	(void)old;

	gtk_widget_set_colormap(widget, cmap);
}

static void cairo_draw_text(cairo_t * cr,
			    PangoFontDescription * desc,
			    GtkRequisition *req,
			    gint padding,
			    const gchar *markup, TextAlign align)
{
	PangoLayout *layout;
	PangoRectangle rect;
	gint x;
	gint y;

	layout = pango_cairo_create_layout(cr);
	pango_layout_set_font_description(layout, desc);
	pango_layout_set_markup(layout, markup, -1);
	pango_layout_get_extents(layout, &rect, NULL);

	switch (align & TEXT_ALIGN_VMASK) {
	case TEXT_ALIGN_BOTTOM:
		y = req->height - padding - PANGO_PIXELS(rect.height);
		break;
	case TEXT_ALIGN_MIDDLE:
		y = (req->height - PANGO_PIXELS(rect.height)) / 2;
		break;
	case TEXT_ALIGN_TOP:
	default:
		y = padding;
		break;
	}
	switch (align & TEXT_ALIGN_HMASK) {
	case TEXT_ALIGN_RIGHT:
		x = req->width - padding - PANGO_PIXELS(rect.width);
		break;
	case TEXT_ALIGN_CENTER:
		x = (req->width - PANGO_PIXELS(rect.width)) / 2;
		break;
	case TEXT_ALIGN_LEFT:
	default:
		x = padding;
		break;
	}

	cairo_save(cr);
	cairo_set_source_rgba(cr, 1.0, 1.0, 1.0, 1.0);
	cairo_move_to(cr, x - 1, y);
	pango_cairo_show_layout(cr, layout);
	cairo_move_to(cr, x + 1, y);
	pango_cairo_show_layout(cr, layout);
	cairo_move_to(cr, x, y - 1);
	pango_cairo_show_layout(cr, layout);
	cairo_move_to(cr, x, y + 1);
	pango_cairo_show_layout(cr, layout);
	cairo_restore(cr);
	cairo_move_to(cr, x, y);
	pango_cairo_show_layout(cr, layout);

	g_object_unref(layout);
}

static void mh_home_sysmon_foreground(HildonDesktopHomeItem *item)
{
	MhHomeSysmonPrivate *priv = MH_HOME_SYSMON_GET_PRIVATE(item);

	priv->foreground = TRUE;
}

static void mh_home_sysmon_background(HildonDesktopHomeItem *item)
{
	MhHomeSysmonPrivate *priv = MH_HOME_SYSMON_GET_PRIVATE(item);

	priv->foreground = FALSE;
}

#ifndef GTKDEBUG
static void mh_home_sysmon_display_ind(DBusGProxy *proxy,
				       gchar *display_state,
				       MhHomeSysmon *sysmon)
{
	MhHomeSysmonPrivate *priv = MH_HOME_SYSMON_GET_PRIVATE(sysmon);

	(void)proxy;

	if (strcmp(display_state, MCE_DISPLAY_OFF_STRING) == 0)
	{
		if (priv->timer_source) {
			g_source_remove(priv->timer_source);
			priv->timer_source = 0;
		}
	} else {
		if (!priv->timer_source) {
			priv->timer_source =
				g_timeout_add(5000,
					      mh_home_sysmon_update, sysmon);
		}
	}
}
#endif
